import os, subprocess, shutil
from rdflib import Graph, BNode
from rdflib.namespace import RDFS
from rdflib import ConjunctiveGraph, URIRef, RDFS, RDF, Namespace
from reqman.apps.reqtool.rest_api.services.parse_ontologies import *
from reqman.apps.reqtool.models.main_req import SYSTEM_CHOICES, FUNCTION_CHOICES, INTERFACE_CHOICES, FLOW_CHOICES, ITEM_CHOICES, STATE_CHOICES, STATE_SET_CHOICES

#HERE
#Ontology_file = "../../Ontologies/Autonomy_v1.ttl"
Ontology_file = "../../Ontologies/2022_AOCS.ttl"

list_of_inference_metrics = ["IncompleteRequirement", "AmbiguousRequirement",
"InconsistentRequirement","NoisyRequirement","OpaqueRequirement","RedundantRequirement"]

prefix_syntax = get_prefix_syntax()

main_syntax = get_main_sytax_inference()

def getclassofprefix(prefix):
    return(findclassofprefix(prefix))

def getclassofmain(main):
    return(findclassofmain(main))

def getinstancefile(data):
    for j in range(0,(len(out_list)-1),2):
        if out_list[j]==data:
            splitter = out_list[j].split(" : ")
            instance_file = out_list[j+1] 
            break
    return instance_file, splitter


def inferencing(project, boilerlate, prefix, main, suffix):
    metrics = {}
    listofboil = []
    listofprefix = []
    listofmain = []
    listofsuffix = []
    for i in boilerlate:
            listofboil.append(list(i.values()))
            #print(list(i.values('title'))[0]['title'])
            metrics[list(i.values('title'))[0]['title']] = []
    for i in prefix:
            listofprefix.append(list(i.values()))
    for i in main:
            listofmain.append(list(i.values()))
    for i in suffix:
            listofsuffix.append(list(i.values()))
    #findclassofmain(listofmain)
    ontology_with_new_req = exportboiltottl(project, listofprefix, listofboil, listofmain, listofsuffix)
    infer_result = shacl(ontology_with_new_req)
    #find in which metrics the requirements belongs e.g. title:REQ-1 is only an Incomplete req (and Ambiguous req), 
    # so -> 'REQ-1' : [Incomplete] ('REQ-1' : [Incomplete, Ambiguous])
    for met in infer_result:
        for title in metrics:
            if title in infer_result[met]:
                metrics[title].append(met)
    metrics.update(infer_result)
    return (metrics, ontology_with_new_req)


def exportboiltottl(project, prefix, boilerplate, main, suffix):
    ''' Export boilerplate to ttl file and return the file path'''
    #print(out_list) #A list with the values and the file that is the instance
    file_name = project.replace('/', '')
    if (os.path.exists("../../Ontologies/"+file_name+"/") == False):
        os.mkdir("../../Ontologies/"+file_name)
    save_path = "../../Ontologies/"+file_name+"/"
    completeName = os.path.join(save_path, file_name)
    per_instances = URIRef("http://delab.csd.auth.gr/ontologies/2018/RDO-instances#")
    rdo = URIRef("http://delab.csd.auth.gr/ontologies/2018/RDO#")
    sao = URIRef("http://delab.csd.auth.gr/ontologies/2018/SAO#")
    rbo = URIRef("http://delab.csd.auth.gr/ontologies/2018/RBO#")
    lo = URIRef("http://delab.csd.auth.gr/ontologies/2018/LO-instances#")
    g = Graph()
    g.load(Ontology_file, format="turtle")
    for i in range(len(boilerplate)):
        title = boilerplate[i][0]['title']
        g.add((per_instances + title, RDF.type, rdo + 'Requirement'))
        if (boilerplate[i][0]['has_prefix'] == True): #If the requirement has Prefix
            prefixmodel=["system_fun_item", "item_function_flow_statevalue"]
            g.add((per_instances + title, rdo + 'hasPrefix', per_instances+(title+"_"+ prefix[i][0]['simple_prefix'])))
            g.add((per_instances+(title+"_"+ prefix[i][0]['simple_prefix']), RDF.type, rbo + prefix[i][0]['simple_prefix']))
            prefix_attr_splitter = prefix[i][0]['item_function_flow_statevalue'].split(" : ")
            if (prefix_attr_splitter[0] == 'StateValue'):
                g.add( (per_instances+(title+"_"+ prefix[i][0]['simple_prefix']), rbo+"isRelatedToLogicalExpression", (per_instances+("SystemStateValueContraint_"+title))))
                g.add((per_instances+("SystemStateValueContraint_"+title), RDF.type, (per_instances+("SystemStateValueContraint"))))
                g.add((per_instances+("SystemStateValueContraint_"+title), rbo+"isRelatedToState", lo + prefix_attr_splitter[1].strip()))
                for j in prefixmodel:
                    if prefix[i][0][j] != "":
                        sub_instance_file, prefix_subject_splitter = getinstancefile(prefix[i][0][j])
                        g.add((per_instances+("SystemStateValueContraint_"+title), rbo+"isRelatedTo"+prefix_subject_splitter[0], URIRef(str(sub_instance_file))+"#"+ str(prefix_subject_splitter[len(prefix_subject_splitter) -1])))
            else:
                prefix_verb_splitter = prefix[i][0]['state_or_verb'].split(" : ")
                g.add((per_instances+(title+"_"+ prefix[i][0]['simple_prefix']), rbo+"isRelatedToOccuringFunctionality", per_instances+(prefix_verb_splitter[0]+"OccuringFunctionality"+title)))
                g.add((per_instances+(prefix_verb_splitter[0]+"OccuringFunctionality"+title), RDF.type, (per_instances+(prefix_verb_splitter[0]+"OccuringFunctionality"))))
                g.add((per_instances+(prefix_verb_splitter[0]+"OccuringFunctionality"+title), rbo+"isRelatedToAction", lo + prefix_verb_splitter[1].strip()))
                for j in prefixmodel:
                    if prefix[i][0][j] != "":
                        sub_instance_file, prefix_subject_splitter = getinstancefile(prefix[i][0][j])
                        g.add((per_instances+(prefix_verb_splitter[0]+"OccuringFunctionality"+title), rbo+"isRelatedTo"+prefix_subject_splitter[0], URIRef(str(sub_instance_file))+"#"+ str(prefix_subject_splitter[len(prefix_subject_splitter) -1])))        
        if (boilerplate[i][0]['has_main'] == True): #If the requirement has Main
            g.add((per_instances + title, rdo + 'hasMain', per_instances+(title+"_"+ main[i][0]['main_choices'])))
            g.add((per_instances+(title+"_"+ main[i][0]['main_choices']), RDF.type, rbo + main[i][0]['main_choices']))
            if main[i][0]['sys_fun_inter'] != "":
                instance_file, splitter_1 = getinstancefile(main[i][0]['sys_fun_inter'])  
                g.add((per_instances+(title+"_"+ main[i][0]['main_choices']), rbo+"isRelatedToSubject", URIRef(str(instance_file))+"#"+ str(splitter_1[len(splitter_1) -1])))
            if main[i][0]['verb'] != "":
                mainverb = main[i][0]['verb'].split(':')
                g.add((per_instances+(title+"_"+ main[i][0]['main_choices']), rbo+"isRelatedToAction", lo + mainverb[1].strip()))
            mainmodel=["flow_function_interface_item_system_state_stateset", "state_item_before_verb", "statevalue_before_verb", "statevalue_system_connection_stateset"]
            for j in mainmodel:
                    if main[i][0][j] != "":
                        instance_file, splitter_1 = getinstancefile(main[i][0][j])  
                        g.add((per_instances+(title+"_"+ main[i][0]['main_choices']), rbo+"isRelatedTo"+splitter_1[0], URIRef(str(instance_file))+"#"+ str(splitter_1[len(splitter_1) -1])))
            if main[i][0]['quantifier'] != "":
                g.add((per_instances+(title+"_"+ main[i][0]['main_choices']), sao+"isRelatedTo", (per_instances+(title+"_"+ (main[i][0]['quantifier']).replace(" ", "")))))
                per = URIRef('http://delab.csd.auth.gr/ontologies/2018/LO#'+ (main[i][0]['quantifier']).replace(" ", ""))
                quant = URIRef(rbo + "Quantifier") # Initial Value if not any
                for s, p, o in g.triples((per, RDFS.subClassOf, None)): #subClassOf
                    print(o)
                    if ('RBO' in o):
                        quant=o
                g.add((per_instances+(title+"_"+ (main[i][0]['quantifier']).replace(" ", "")), RDF.type, quant))
                g.add((per_instances+(title+"_"+ (main[i][0]['quantifier']).replace(" ", "")),  rbo+"hasAdverb", lo + (main[i][0]['quantifier']).replace(" ", "")))
                if main[i][0]['numerical'] != "":
                    g.add((per_instances+(title+"_"+ (main[i][0]['quantifier']).replace(" ", "")),  rbo+"hasNumerical", URIRef(per_instances+(main[i][0]['quantifier']).replace(" ", "")+"_"+str(main[i][0]['numerical']))))
                    g.add((URIRef(per_instances+(main[i][0]['quantifier']).replace(" ", "")+"_"+str(main[i][0]['numerical'])),  RDF.type, URIRef(rbo+"Numerical")))
        if (boilerplate[i][0]['has_suffix'] == True):
            g.add((per_instances + title, rdo + 'hasSuffix', per_instances+(title+"_"+ suffix[i][0]['suffix_choices'])))
            #Need S1
            #Need S2
            #Need S3
            g.add((per_instances+(title+"_"+ suffix[i][0]['suffix_choices']), RDF.type, URIRef(rbo+ suffix[i][0]['suffix_choices'])))
        
        
    g.serialize(destination=save_path+file_name+'.ttl', format='turtle')
    return (save_path+file_name)
    #print(g.serialize(format="turtle").decode("utf-8"))

def shacl(ontotlogy_file):
    metrics_dict = {}
    file1 = ontotlogy_file + ".ttl"
    target = ontotlogy_file+"infer.ttl"
    open(target, 'w').close()
    shutil.copyfile(file1, target)
    for i in range(3):
            print(i+1, 'Cycle of inferencing')
            result = subprocess.check_output(["./reqman/apps/reqtool/rest_api/services/shacl-1.3.2/bin/shaclinfer.sh", "-datafile", target])
            file_object = open(target, 'a')
            res = str(result,'utf-8')
            file_object.write(res.strip())
            file_object.close()
    
    #need to fiil with all the metrics
    g = Graph()
    g.load(target, format="turtle")
    for inf_metr in list_of_inference_metrics:
        metrics_dict[inf_metr] = []
        p = URIRef("http://delab.csd.auth.gr/ontologies/2018/RDO#"+inf_metr)
        for s, p, o in g.triples((None, RDF.type, p)):
            #metrics_dict["IncompleteRequirement"] = [s]
            get_title = str(s).split("#")
            metrics_dict[inf_metr].append(get_title[1])
    print(metrics_dict)
    return(metrics_dict)

'''Find if the Prefix exists in syntax'''
def findclassofprefix(listofprefix):
    classlist=[]
    currentverb = listofprefix['state_or_verb'].rsplit(' : ', 1)
    currentsubject = listofprefix['system_fun_item']
    currentattr = listofprefix['item_function_flow_statevalue']
    for prefix in prefix_syntax:
        for inprefix in prefix_syntax[prefix]:
            if currentverb[1] in prefix_syntax[prefix][inprefix]:
                if currentsubject.split(":")[0].strip() in prefix_syntax[prefix][inprefix.split("_")[0]+"_Subject"]:
                    if currentattr.split(":")[0].strip()  in prefix_syntax[prefix][inprefix.split("_")[0]+"_Attribute"]:
                        if listofprefix['prefix'] == "If":
                            return "P1"
                        if listofprefix['prefix'] == "As soon as":
                            return "P2"
                        if listofprefix['prefix'] == "As long as":
                            return "P3"
                        if listofprefix['prefix'] == "Once":
                            return "P1"
                        if listofprefix['prefix'] == "Unless":
                            return "P1"
                        return(prefix)
    classlist.append("ERROR")
    return classlist



'''Find if the Main exists in syntax'''
def findclassofmain(listofmain):
    classlist = []
    for i in range(len(main_syntax)):
        verbsmain =  main_syntax[('M'+str(i+1))]['Verbs']
        subsmain =  main_syntax[('M'+str(i+1))]['Subject']
        attrsmain =  main_syntax[('M'+str(i+1))]['Attributes']
        #for currentmain in range(len(listofmain)):
        currentverb = listofmain['verb'].rsplit(' : ', 1)
        currentsubject = listofmain['sys_fun_inter']
        currentattr = listofmain['flow_function_interface_item_system_state_stateset']
        for verbmain in verbsmain:
            flag = 0
            if (verbmain.upper() == currentverb[len(currentverb)-1].upper() ):
                for submain in subsmain:
                    flag = 0
                    if submain == "System":
                        for sys in SYSTEM_CHOICES:
                            if sys[0] ==  currentsubject:
                                flag = 1
                                break
                    if submain == "Function":
                        for sys in FUNCTION_CHOICES:
                            if sys[0] ==  currentsubject:
                                flag = 1
                                break
                    if submain == "Interface":
                        for sys in INTERFACE_CHOICES:
                            if sys[0] ==  currentsubject:
                                flag = 1    
                                break
                    if flag == 1 :
                        for attrmain in attrsmain:
                            if attrmain == "Flow":
                                for sys in FLOW_CHOICES:
                                    if sys[0] ==  currentattr:
                                        classlist.append('M'+str(i+1))
                                        return(classlist)
                            if attrmain == "Item":
                                for sys in ITEM_CHOICES:
                                    if sys[0] ==  currentattr:
                                        classlist.append('M'+str(i+1))
                                        return(classlist)
                            if attrmain == "Function":
                                for sys in FUNCTION_CHOICES:
                                    if sys[0] ==  currentattr:
                                        classlist.append('M'+str(i+1))
                                        return(classlist)
                            if attrmain == "System":
                                for sys in SYSTEM_CHOICES:
                                    if sys[0] ==  currentattr:
                                        classlist.append('M'+str(i+1))
                                        return(classlist)
                            if attrmain == "Interface":
                                for sys in INTERFACE_CHOICES:
                                    if sys[0] ==  currentattr:
                                        classlist.append('M'+str(i+1))
                                        return(classlist)
                            if attrmain == "State":
                                for sys in (STATE_CHOICES):
                                    if sys[0] ==  currentattr:
                                        classlist.append('M'+str(i+1))
                                        return(classlist)
                            if attrmain == "StateSet":
                                for sys in (STATE_SET_CHOICES):
                                    if sys[0] ==  currentattr:
                                        classlist.append('M'+str(i+1))
                                        return(classlist)
            
    classlist.append("ERROR")
    return classlist
                              

